// vidcapmodule.cpp
// by Markus Gritsch (gritsch@iue.tuwien.ac.at)

// code plugged together from
// http://msdn.microsoft.com/library/psdk/directx/dx8_c/ds/AppDev/Capture_Intro.htm
// http://msdn.microsoft.com/library/psdk/directx/dx8_c/ds/AppDev/Using_SampleGrab.htm
// http://msdn.microsoft.com/library/psdk/directx/dx8_c/ds/Samples/PlayCap.htm
// http://www.python.org/doc/current/ext/ext.html


#include "Python.h"

#include <dshow.h>
#include <qedit.h> // Sample Grabber, Null Renderer


static PyObject *VidcapError;

staticforward PyTypeObject Dev_Type;


typedef struct {
    PyObject_HEAD

    IGraphBuilder         *ob_pGraph;
    ICaptureGraphBuilder2 *ob_pBuilder;
    IBaseFilter           *ob_pSrc;
    ISampleGrabber        *ob_pGrab;
    IMediaControl         *ob_pControl;
    //IMediaEvent           *ob_pEvent;
    IMediaEventEx         *ob_pEvent;
} DevObject;


static void
cleanup(DevObject *self)
{
    if (self->ob_pControl != NULL)
        self->ob_pControl->Stop();

    if (self->ob_pEvent != NULL)
        self->ob_pEvent->Release();
    if (self->ob_pControl != NULL)
        self->ob_pControl->Release();
    if (self->ob_pGrab != NULL)
        self->ob_pGrab->Release();
    if (self->ob_pSrc != NULL)
        self->ob_pSrc->Release();
    if (self->ob_pBuilder != NULL)
        self->ob_pBuilder->Release();
    if (self->ob_pGraph != NULL)
        self->ob_pGraph->Release();
}


static DevObject *
newDevObject(PyObject *args)
{
    DevObject *self;
    self = PyObject_New(DevObject, &Dev_Type);
    if (self == NULL)
        return NULL;

    self->ob_pGraph   = NULL;
    self->ob_pBuilder = NULL;
    self->ob_pSrc     = NULL;
    self->ob_pGrab    = NULL;
    self->ob_pControl = NULL;
    self->ob_pEvent   = NULL;


    int devnum;
    int showVideoWindow;

    if (!PyArg_ParseTuple(args, "ii:newDev", &devnum, &showVideoWindow))
        return NULL;

    HRESULT hr;


    // Creating the Required Components
    // ================================

    // Create the filter graph.
    hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC,
                          IID_IGraphBuilder, (void **)&self->ob_pGraph);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Creation of the filter graph failed.");
        cleanup(self);
        return NULL;
    }

    // Create the capture graph builder.
    hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC,
                          IID_ICaptureGraphBuilder2, (void **)&self->ob_pBuilder);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Creation of the capture graph builder failed.");
        cleanup(self);
        return NULL;
    }

    hr = self->ob_pBuilder->SetFiltergraph(self->ob_pGraph);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Setting the filter graph for the capture graph builder failed.");
        cleanup(self);
        return NULL;
    }


    // Selecting a Capture Device
    // ==========================

    // Create the system device enumerator.
    ICreateDevEnum *pDevEnum = NULL;
    hr = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC,
                          IID_ICreateDevEnum, (void **)&pDevEnum);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Creation of the system device enumerator failed.");
        cleanup(self);
        return NULL;
    }

    // Create an enumerator for video capture devices.
    IEnumMoniker *pClassEnum = NULL;
    hr = pDevEnum->CreateClassEnumerator(CLSID_VideoInputDeviceCategory, &pClassEnum, 0);
    pDevEnum->Release();
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Creation of the enumerator for video capture devices failed.");
        cleanup(self);
        return NULL;
    }

    ULONG cFetched;
    IMoniker *pMoniker = NULL;
    for (int i=0; i <= devnum; i++)
    {
        if (pClassEnum->Next(1, &pMoniker, &cFetched) != S_OK)
        {
            PyErr_SetString(VidcapError, "The device with the specified device number does not exist.");
            pClassEnum->Release();
            cleanup(self);
            return NULL;
        }
    }
    pClassEnum->Release();

    // Bind the moniker to a filter object.
    hr = pMoniker->BindToObject(0, 0, IID_IBaseFilter, (void**)&self->ob_pSrc);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Binding the moniker to a filter object failed.");
        cleanup(self);
        return NULL;
    }
    pMoniker->Release();


    // Now add the capture filter to the graph
    hr = self->ob_pGraph->AddFilter(self->ob_pSrc, L"Video Capture");
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Adding the capture filter to the filter graph failed.\n              Device maybe not connected or already in use.");
        cleanup(self);
        return NULL;
    }


    // Adding the Sample Grabber to the Filter Graph
    // =============================================

    IBaseFilter     *pF = NULL;
    AM_MEDIA_TYPE   mt;

    hr = CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC,
                          IID_IBaseFilter, (LPVOID *)&pF);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Creation of the sample grabber failed.");
        cleanup(self);
        return NULL;
    }

    hr = pF->QueryInterface(IID_ISampleGrabber, (void **)&self->ob_pGrab);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Sample grabber interface could not be found.");
        cleanup(self);
        return NULL;
    }

    hr = self->ob_pGraph->AddFilter(pF, L"Grabber");
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Adding the sample grabber to the filter graph failed.");
        cleanup(self);
        return NULL;
    }

    ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
    mt.majortype = MEDIATYPE_Video;
    mt.subtype = MEDIASUBTYPE_RGB24;
    mt.formattype = FORMAT_VideoInfo;
    hr = self->ob_pGrab->SetMediaType(&mt);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Setting the media type of the sample grabber failed.");
        cleanup(self);
        return NULL;
    }


    // Rendering the Streams
    // =====================

    hr = self->ob_pBuilder->RenderStream(
            &PIN_CATEGORY_PREVIEW,
            &MEDIATYPE_Video,
            self->ob_pSrc, // src
            pF,            // via
            NULL           // dest
        );
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Capture Graph could not be created.");
        cleanup(self);
        return NULL;
    }

    pF->Release();


    // Controlling a Capture Graph
    // ===========================

    if (!showVideoWindow)
    {
        IVideoWindow *pWindow;

        hr = self->ob_pGraph->QueryInterface(IID_IVideoWindow, (void **)&pWindow);
        if (FAILED(hr))
        {
            PyErr_SetString(VidcapError, "Video Window interface could not be found.");
            cleanup(self);
            return NULL;
        }

        hr = pWindow->put_AutoShow(OAFALSE);
        if (FAILED(hr))
        {
            PyErr_SetString(VidcapError, "Video Window hiding failed.");
            cleanup(self);
            return NULL;
        }

        pWindow->Release();
    }

    hr = self->ob_pGraph->QueryInterface(IID_IMediaControl, (void **)&self->ob_pControl);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Media control interface could not be found.");
        cleanup(self);
        return NULL;
    }

    hr = self->ob_pGraph->QueryInterface(IID_IMediaEvent, (void **)&self->ob_pEvent);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Media event interface could not be found.");
        cleanup(self);
        return NULL;
    }


    // Set up one-shot mode.
    hr = self->ob_pGrab->SetBufferSamples(TRUE);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Setting up buffering of samples failed.");
        cleanup(self);
        return NULL;
    }

    hr = self->ob_pGrab->SetOneShot(TRUE);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Setting up one-shot mode of the sample grabber failed.");
        cleanup(self);
        return NULL;
    }


    return self;
}


static void
Dev_dealloc(DevObject *self)
{
    cleanup(self);

    PyObject_Del(self);
}


static PyObject *
Dev_getbuffer(DevObject *self, PyObject *args)
{
    long timeout;

    if (!PyArg_ParseTuple(args, "l:getbuffer", &timeout))
        return NULL;
    if (timeout == 0)
        timeout = INFINITE;

    HRESULT hr;

    long evCode;
/*
            long evParam1, evParam2;
            while (SUCCEEDED(self->ob_pEvent->GetEvent(&evCode, &evParam1, &evParam2, 0)))
            {
                // Handle the event (not shown).
                hr = self->ob_pEvent->FreeEventParams(evCode, evParam1, evParam2);
            }
*/

/*
            hr = self->ob_pEvent->SetNotifyFlags(AM_MEDIAEVENT_NONOTIFY);
*/
    hr = self->ob_pControl->Run();
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Running the capture graph failed.");
        cleanup(self);
        return NULL;
    }
/*
            OAFilterState pfs;
            hr = self->ob_pControl->GetState(2000, &pfs);
*/

/*
            HANDLE  hEvent;
            long    param1, param2;

            hr = self->ob_pEvent->GetEventHandle((OAEVENT*)&hEvent);

            if (WAIT_OBJECT_0 == WaitForSingleObject(hEvent, 0))
            {
    	        while (hr = self->ob_pEvent->GetEvent(&evCode, &param1, &param2, 0), SUCCEEDED(hr))
                {
                    printf("Event code: %#04x\n    Params: %d, %d\n", evCode, param1, param2);
                    hr = self->ob_pEvent->FreeEventParams(evCode, param1, param2);
                }
            }
*/

    hr = self->ob_pEvent->WaitForCompletion(timeout, &evCode);
    // *VERY* strange: Sometimes, the EC_COMPLETE event doesn't occure,
    // and WaitForCompletion() only returns after the timeout interval.
    // Because of this strage behaviour the following if(FAILED(hr))
    // is commented out.  Any suggestions would be very welcome!
/*    if (FAILED(hr))
    {
        switch (hr)
        {
        case E_ABORT:
            PyErr_SetString(VidcapError, "Capture not performed (timeout expired).");
            break;
        case VFW_E_WRONG_STATE:
            PyErr_SetString(VidcapError, "Capture not performed (filter graph not running).");
            break;
        default:
            PyErr_SetString(VidcapError, "UNKNOWN CASE!!");
        }
        return NULL;
    }
*/
    hr = self->ob_pControl->Stop();
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Stopping the capture graph failed.");
        return NULL;
    }

    AM_MEDIA_TYPE MediaType;
    hr = self->ob_pGrab->GetConnectedMediaType(&MediaType);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Getting the sample grabber's connected media type failed.");
        return NULL;
    }

    // Get a pointer to the video header.
    VIDEOINFOHEADER *pVideoHeader = (VIDEOINFOHEADER*)MediaType.pbFormat;

    long size   = pVideoHeader->bmiHeader.biSizeImage;
    long width  = pVideoHeader->bmiHeader.biWidth;
    long height = pVideoHeader->bmiHeader.biHeight;

    // Allocate memory.
    //void *buffer = malloc(size);
    void *buffer = PyMem_Malloc(size);
    if (!buffer)
    {
    PyErr_NoMemory();
        return NULL;
    }

    // Copy the image into the buffer.
    hr = self->ob_pGrab->GetCurrentBuffer(&size, (long *)buffer);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Getting the sample grabber's current buffer failed.");
        return NULL;
    }

    // Return the buffer.
    PyObject *value;
    value = Py_BuildValue("(s#,l,l)", buffer, size, width, height);

    //free(buffer);
    PyMem_Free(buffer);

    return value;
}


static PyMethodDef Dev_methods[] = {
    {"getbuffer",   (PyCFunction)Dev_getbuffer,  METH_VARARGS},
    {NULL,          NULL}           /* sentinel */
};


static PyObject *
Dev_getattr(DevObject *self, char *name)
{
    return Py_FindMethod(Dev_methods, (PyObject *)self, name);
}


statichere PyTypeObject Dev_Type = {
    /* The ob_type field must be initialized in the module init function
     * to be portable to Windows without using C++. */
    PyObject_HEAD_INIT(NULL)
    0,                        /*ob_size*/
    "Dev",                    /*tp_name*/
    sizeof(DevObject),        /*tp_basicsize*/
    0,                        /*tp_itemsize*/
    /* methods */             
    (destructor)Dev_dealloc,  /*tp_dealloc*/
    0,                        /*tp_print*/
    (getattrfunc)Dev_getattr, /*tp_getattr*/
    (setattrfunc)0,           /*tp_setattr*/
    0,                        /*tp_compare*/
    0,                        /*tp_repr*/
    0,                        /*tp_as_number*/
    0,                        /*tp_as_sequence*/
    0,                        /*tp_as_mapping*/
    0,                        /*tp_hash*/
};


static PyObject *
vidcap_new_Dev(PyObject *self, PyObject *args)
{
    DevObject *rv;
    
    rv = newDevObject(args);
    if ( rv == NULL )
        return NULL;
    return (PyObject *)rv;
}


static PyMethodDef vidcap_methods[] = {
    {"new_Dev", vidcap_new_Dev, METH_VARARGS},
    {NULL,      NULL}        /* Sentinel */
};


extern "C"

/* platform independent */
#ifdef MS_WIN32
__declspec(dllexport)
#endif

void
initvidcap(void)
{
    Dev_Type.ob_type = &PyType_Type;

    PyObject *m, *d;

    m = Py_InitModule("vidcap", vidcap_methods);
    d = PyModule_GetDict(m);
    VidcapError = PyErr_NewException("vidcap.Error", NULL, NULL);
    PyDict_SetItemString(d, "error", VidcapError);

    HRESULT hr;

    // Initialize COM.
    hr = CoInitialize(NULL);
    if (FAILED(hr))
    {
        PyErr_SetString(VidcapError, "Initialization of COM failed.");
        return;
    }

    // Register Release COM at exit.
    if ( Py_AtExit( (void (__cdecl *)(void))CoUninitialize ) )
    {
        PyErr_SetString(VidcapError, "Registering CoUninitialize at exit failed.");
        CoUninitialize();
        return;
    }
}
